@@ -1,5 +1,5 @@
 /*
-  Copyright (c) 2016 MariaDB Corporation
+  Copyright (c) 2016, 2019, MariaDB Corporation.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
   All methods pertaining to temporary tables.
 */
 
+#include "mariadb.h"
 #include "sql_acl.h"                            /* TMP_TABLE_ACLS */
 #include "sql_base.h"                           /* tdc_create_key */
 #include "lock.h"                               /* mysql_lock_remove */
@@ -64,7 +65,8 @@ TABLE *THD::create_and_open_tmp_table(handlerton *hton,
                                       const char *path,
                                       const char *db,
                                       const char *table_name,
-                                      bool open_in_engine)
+                                      bool open_in_engine,
+                                      bool open_internal_tables)
 {
   DBUG_ENTER("THD::create_and_open_tmp_table");
 
@@ -73,7 +75,9 @@ TABLE *THD::create_and_open_tmp_table(handlerton *hton,
 
   if ((share= create_temporary_table(hton, frm, path, db, table_name)))
   {
+    open_options|= HA_OPEN_FOR_CREATE;
     table= open_temporary_table(share, table_name, open_in_engine);
+    open_options&= ~HA_OPEN_FOR_CREATE;
 
     /*
       Failed to open a temporary table instance. As we are not passing
@@ -87,6 +91,15 @@ TABLE *THD::create_and_open_tmp_table(handlerton *hton,
 
       /* Free the TMP_TABLE_SHARE. */
       free_tmp_table_share(share, false);
+      DBUG_RETURN(0);
+    }
+
+    /* Open any related tables */
+    if (open_internal_tables && table->internal_tables &&
+        open_and_lock_internal_tables(table, open_in_engine))
+    {
+      drop_temporary_table(table, NULL, false);
+      DBUG_RETURN(0);
     }
   }
 
@@ -248,7 +261,7 @@ TMP_TABLE_SHARE *THD::find_tmp_table_share(const TABLE_LIST *tl)
   @return Success                     A pointer to table share object
           Failure                     NULL
 */
-TMP_TABLE_SHARE *THD::find_tmp_table_share(const char *key, uint key_length)
+TMP_TABLE_SHARE *THD::find_tmp_table_share(const char *key, size_t key_length)
 {
   DBUG_ENTER("THD::find_tmp_table_share");
 
@@ -313,7 +326,7 @@ TMP_TABLE_SHARE *THD::find_tmp_table_share(const char *key, uint key_length)
 bool THD::open_temporary_table(TABLE_LIST *tl)
 {
   DBUG_ENTER("THD::open_temporary_table");
-  DBUG_PRINT("enter", ("table: '%s'.'%s'", tl->db, tl->table_name));
+  DBUG_PRINT("enter", ("table: '%s'.'%s'", tl->db.str, tl->table_name.str));
 
   TMP_TABLE_SHARE *share;
   TABLE *table= NULL;
@@ -338,7 +351,7 @@ bool THD::open_temporary_table(TABLE_LIST *tl)
     DBUG_RETURN(false);
   }
 
-  if (!tl->db)
+  if (!tl->db.str)
   {
     DBUG_PRINT("info",
                ("Table reference to a temporary table must have database set"));
@@ -399,7 +412,7 @@ bool THD::open_temporary_table(TABLE_LIST *tl)
     if (tl->open_type == OT_TEMPORARY_ONLY &&
         tl->open_strategy == TABLE_LIST::OPEN_NORMAL)
     {
-      my_error(ER_NO_SUCH_TABLE, MYF(0), tl->db, tl->table_name);
+      my_error(ER_NO_SUCH_TABLE, MYF(0), tl->db.str, tl->table_name.str);
       DBUG_RETURN(true);
     }
     DBUG_RETURN(false);
@@ -553,25 +566,21 @@ bool THD::close_temporary_tables()
           true                        Error
 */
 bool THD::rename_temporary_table(TABLE *table,
-                                 const char *db,
-                                 const char *table_name)
+                                 const LEX_CSTRING *db,
+                                 const LEX_CSTRING *table_name)
 {
-  DBUG_ENTER("THD::rename_temporary_table");
-
   char *key;
   uint key_length;
-
   TABLE_SHARE *share= table->s;
+  DBUG_ENTER("THD::rename_temporary_table");
 
   if (!(key= (char *) alloc_root(&share->mem_root, MAX_DBKEY_LENGTH)))
-  {
     DBUG_RETURN(true);
-  }
 
   /*
     Temporary tables are renamed by simply changing their table definition key.
   */
-  key_length= create_tmp_table_def_key(key, db, table_name);
+  key_length= create_tmp_table_def_key(key, db->str, table_name->str);
   share->set_table_cache_key(key, key_length);
 
   DBUG_RETURN(false);
@@ -689,19 +698,20 @@ bool THD::rm_temporary_table(handlerton *base, const char *path)
   char frm_path[FN_REFLEN + 1];
 
   strxnmov(frm_path, sizeof(frm_path) - 1, path, reg_ext, NullS);
-  if (mysql_file_delete(key_file_frm, frm_path, MYF(0)))
-  {
-    error= true;
-  }
-  file= get_new_handler((TABLE_SHARE*) 0, current_thd->mem_root, base);
+
+  file= get_new_handler((TABLE_SHARE*) 0, mem_root, base);
   if (file && file->ha_delete_table(path))
   {
     error= true;
     sql_print_warning("Could not remove temporary table: '%s', error: %d",
                       path, my_errno);
   }
-
   delete file;
+
+  if (mysql_file_delete(key_file_frm, frm_path, MYF(0)))
+  {
+    error= true;
+  }
   DBUG_RETURN(error);
 }
 
@@ -866,7 +876,7 @@ void THD::restore_tmp_table_share(TMP_TABLE_SHARE *share)
   @return false                       Temporary tables exist
           true                        No temporary table exist
 */
-inline bool THD::has_temporary_tables()
+bool THD::has_temporary_tables()
 {
   DBUG_ENTER("THD::has_temporary_tables");
   bool result= (rgi_slave
@@ -1065,7 +1075,7 @@ TABLE *THD::find_temporary_table(const char *key, uint key_length,
         case TMP_TABLE_ANY:        found= true;                 break;
         }
       }
-      if (table && unlikely(table->m_needs_reopen))
+      if (table && unlikely(table->needs_reopen()))
       {
         share->all_tmp_tables.remove(table);
         free_temporary_table(table);
@@ -1099,22 +1109,25 @@ TABLE *THD::find_temporary_table(const char *key, uint key_length,
           Failure                     NULL
 */
 TABLE *THD::open_temporary_table(TMP_TABLE_SHARE *share,
-                                 const char *alias,
+                                 const char *alias_arg,
                                  bool open_in_engine)
 {
+  TABLE *table;
+  LEX_CSTRING alias= {alias_arg, strlen(alias_arg) };
   DBUG_ENTER("THD::open_temporary_table");
 
-  TABLE *table;
 
   if (!(table= (TABLE *) my_malloc(sizeof(TABLE), MYF(MY_WME))))
   {
     DBUG_RETURN(NULL);                          /* Out of memory */
   }
 
-  if (open_table_from_share(this, share, alias,
+  if (open_table_from_share(this, share, &alias,
                             open_in_engine ? (uint)HA_OPEN_KEYFILE : 0,
-                            EXTRA_RECORD, open_options | ha_open_options, table,
-                            open_in_engine ? false : true))
+                            EXTRA_RECORD,
+                            (ha_open_options |
+                             (open_options & HA_OPEN_FOR_CREATE)),
+                            table, open_in_engine ? false : true))
   {
     my_free(table);
     DBUG_RETURN(NULL);
@@ -1122,8 +1135,9 @@ TABLE *THD::open_temporary_table(TMP_TABLE_SHARE *share,
 
   table->reginfo.lock_type= TL_WRITE;           /* Simulate locked */
   table->grant.privilege= TMP_TABLE_ACLS;
-  share->tmp_table= (table->file->has_transactions() ?
+  share->tmp_table= (table->file->has_transaction_manager() ?
                      TRANSACTIONAL_TMP_TABLE : NON_TRANSACTIONAL_TMP_TABLE);
+  share->not_usable_by_query_cache= 1;
 
   table->pos_in_table_list= 0;
   table->query_id= query_id;
@@ -1137,7 +1151,8 @@ TABLE *THD::open_temporary_table(TMP_TABLE_SHARE *share,
     thread_safe_increment32(&slave_open_temp_tables);
   }
 
-  DBUG_PRINT("tmptable", ("Opened table: '%s'.'%s'%p", table->s->db.str,
+  DBUG_PRINT("tmptable", ("Opened table: '%s'.'%s  table: %p",
+                          table->s->db.str,
                           table->s->table_name.str, table));
   DBUG_RETURN(table);
 }
@@ -1373,15 +1388,15 @@ bool THD::log_events_and_free_tmp_shares()
           /*
              We are going to add ` around the table names and possible more
              due to special characters.
-           */
-          append_identifier(this, &s_query, share->table_name.str,
-              share->table_name.length);
+          */
+          append_identifier(this, &s_query, &share->table_name);
           s_query.append(',');
         }
         rm_temporary_table(share->db_type(), share->path.str);
         free_table_share(share);
         my_free(share);
       }
+
       if (at_least_one_create_logged)
       {
         clear_error();
@@ -1398,7 +1413,8 @@ bool THD::log_events_and_free_tmp_shares()
 
         get_stmt_da()->set_overwrite_status(true);
         transaction.stmt.mark_dropped_temp_table();
-        if ((error= (mysql_bin_log.write(&qinfo) || error)))
+        bool error2= mysql_bin_log.write(&qinfo);
+        if (unlikely(error|= error2))
         {
           /*
              If we're here following THD::cleanup, thence the connection
@@ -1411,7 +1427,7 @@ bool THD::log_events_and_free_tmp_shares()
              that THD::close_tables failed. (Actually, the SQL
              thread only calls THD::close_tables while applying
              old Start_log_event_v3 events.)
-           */
+          */
           sql_print_error("Failed to write the DROP statement for "
               "temporary tables to binary log");
         }
